"""A module containing functions to run MCMC using emcee."""

import numpy as np
import emcee
from eztao.ts import carma_fit, neg_param_ll, neg_fcoeff_ll
from eztao.carma import CARMA_term
from celerite import GP


def mcmc(t, y, yerr, p, q, n_walkers=32, burn_in=500, n_samples=2000, init_param=None):
    """
    A simple wrapper to run quick MCMC using emcee.

    Args:
        t (array(float)): Time stamps of the input time series (the default unit is day).
        y (array(float)): y values of the input time series.
        yerr (array(float)): Measurement errors for y values.
        p (int): The p order of a CARMA(p, q) model.
        q (int): The q order of a CARMA(p, q) model.
        n_walkers (int, optional): Number of MCMC walkers. Defaults to 32.
        burn_in (int, optional): Number of burn in steps. Defaults to 500.
        n_samples (int, optional): Number of MCMC steps to run. Defaults to 2000.
        init_param (array(float), optional): The initial position for the MCMC walker.
            Defaults to None.

    Returns:
        (object, array(float), array(float)): The emcee sampler object. The MCMC
        flatchain (n_walkers*n_samplers, dim) and chain (n_walkers, n_samplers, dim)
        in CARMA space if p > 2, otherwise empty.
    """
    assert p > q, "p order must be greater than q order."

    if init_param is not None and p <= 2:
        assert len(init_param) == int(
            p + q + 1
        ), "The initial parameters doesn't match the dimension of the CARMA model!"
    else:
        print("Searching for best-fit CARMA parameters...")
        init_param = carma_fit(t, y, yerr, p, q, n_opt=200)

    # set on param or fcoeff
    if p > 2:
        ll = lambda *args: -neg_fcoeff_ll(*args)
        init_sample = CARMA_term.carma2fcoeffs_log(
            np.log(init_param[:p]), np.log(init_param[p:])
        )
        init_sample = np.exp(init_sample)
    else:
        ll = lambda *args: -neg_param_ll(*args)
        init_sample = init_param

    # create vectorized functions
    vec_fcoeff2carma_log = np.vectorize(
        CARMA_term.fcoeffs2carma_log,
        excluded=[
            1,
        ],
        signature="(n)->(m),(k)",
    )

    # reposition ts
    t = t - t[0]
    y = y - np.median(y)

    # init celerite kernel/GP
    kernel = CARMA_term(np.log(init_param[:p]), np.log(init_param[p:]))
    gp = GP(kernel, mean=0)
    gp.compute(t, yerr)

    # init sampler
    ndim = len(init_sample)
    sampler = emcee.EnsembleSampler(n_walkers, ndim, ll, args=[y, gp])

    print("Running burn-in...")
    rng = np.random.default_rng()
    p0 = np.log(init_sample) + 1e-8 * rng.standard_normal((n_walkers, ndim))
    p0, lp, _ = sampler.run_mcmc(p0, burn_in)

    print("Running production...")
    sampler.reset()
    sampler.run_mcmc(p0, n_samples)

    carma_flatchain = np.array([])
    carma_chain = np.array([])

    if p > 2:
        logAR, logMA = vec_fcoeff2carma_log(sampler.flatchain, p)
        carma = np.hstack((logAR, logMA))
        carma_flatchain = carma
        carma_chain = carma.reshape((n_walkers, n_samples, ndim), order="F")

    return sampler, carma_flatchain, carma_chain



"""
A collection of utility functions to assist analysis/simulation of time series data.
"""

import numpy as np
from functools import partial
from scipy.stats import median_abs_deviation as mad

__all__ = ["downsample_byN", "add_season", "downsample_byTime", "median_clip"]


def downsample_byN(t, nObs, seed):
    """
    Randomly choose N data points from a given time series.

    Args:
        t (array(float)): Time stamps of the original time series.
        N (int): The number of entries in the final time series.
        seed (int): Random seed for downsampling. Defaults to None.

    Returns:
        A 1d array booleans indicating which data points to keep.
    """

    if seed is not None:
        rng = np.random.default_rng(seed=seed)
    else:
        rng = np.random.default_rng()

    # random choose index
    idx = np.arange(len(t))
    mask = np.zeros_like(idx, dtype=bool)
    true_idx = rng.choice(idx, nObs, replace=False)

    # assign chosen index to 1/True
    mask[true_idx] = 1

    return mask


def _get_nearest_idx(tIn, x):
    """Internal function to return nearest neighbor in an array."""
    return (np.abs(tIn - x)).argmin()


def downsample_byTime(tIn, tOut):
    """
    Downsample a time series at desired output time stamps.

    Args:
        tIn (array(float)): Time stamps of the original time series.
        tOut (array(float)): Time stamps of the output time series.

    Returns:
        array(int): Indices for which the data points should be kept from the original
        time series. Note that there could be duplicates.
    """
    get_nearest = partial(_get_nearest_idx, tIn)
    return np.array(list(map(get_nearest, tOut)))


def add_season(t, lc_start=0, season_start=90, season_end=270):
    """
    Insert seasonal gaps into time series

    Args:
        t (array(float)): Time stamps of the original time series.
        lc_start (float): Starting day for the output time series. (0 -> 365.25).
            Default to 0.
        season_start (float): Observing season start day within a year. Default to 90.
        season_end (float): Observing season end day within a year. Default to 270.

    Returns:
        A 1d array booleans indicating which data points to keep.
    """
    t = t - t[0]
    t = t + lc_start

    mask = (np.mod(t, 365.25) > season_start) & (np.mod(t, 365.25) < season_end)

    return mask


def median_clip(y, num_sigma=3):
    r"""Clip time series using a three point median filter.

    The sigma (standard deviation) for the time series is computed from the median  absolute deviation (MAD) as to reduce the effects from extreme outliers, where
    sigma :math:`\sim` 1.4826*MAD. If more than 10% of the data points are removed, the upper
    bound will be lifted gradually until that fraction drops bellow 10%.

    Args:
        y (array(float)): y values of the original time series.
        num_sigma (int, optional): Data points that are more than this number of sigma
            away from the three point median will be removed. Defaults to 3.

    Returns:
        A 1d array booleans indicating which data points to keep.
    """
    y = np.atleast_1d(y)
    lc_len = y.shape[0]
    med = y.copy()
    med[1:-2] = [np.median(y[i - 1 : i + 2]) for i in range(1, lc_len - 2)]

    # need to deal with edges of median filter
    med[0] = np.median(np.array([y[-1], y[0], y[1]]))
    med[-1] = np.median(np.array([y[-2], y[-1], y[0]]))

    # compute residual & sigma
    res = np.abs(y - med)
    sigma = mad(res, scale="normal")

    # set clipping thresh hold
    raise_bar = True
    thresh = num_sigma * sigma

    # if remove too much, raise bar until only remove 10%
    while raise_bar:
        ratio = np.sum(res > thresh) / lc_len
        if ratio < 0.1:
            break
        else:
            thresh += 0.1 * sigma
    return res < thresh



"""
A collection of functions for simulating CARMA processes.
"""

import numpy as np
from math import ceil
import celerite
from celerite import GP
from eztao.ts.utils import add_season, downsample_byN, downsample_byTime
from eztao.carma.CARMATerm import DRW_term, CARMA_term

__all__ = ["gpSimFull", "gpSimRand", "gpSimByTime", "addNoise", "pred_lc"]


def gp_sample(gp, size=None, seed=None):
    """Sample Celerite GP with a fixed seed"""

    if seed is not None:
        rng = np.random.default_rng(seed=seed)
    else:
        rng = np.random.default_rng()

    gp._recompute()
    if size is None:
        n = rng.standard_normal(len(gp._t))
    else:
        n = rng.standard_normal((len(gp._t), size))

    n = gp.solver.dot_L(n)
    if size is None:
        return gp.mean.get_value(gp._t) + n[:, 0]
    return gp.mean.get_value(gp._t)[None, :] + n.T


def gpSimFull(carmaTerm, SNR, duration, N, nLC=1, log_flux=True, lc_seed=None):
    """
    Simulate CARMA time series using uniform sampling.

    Args:
        carmaTerm (object): An EzTao CARMA kernel.
        SNR (float): Signal-to-noise defined as ratio between CARMA RMS amplitude and
            the median of the measurement errors (simulated using log normal).
        duration (float): The duration of the simulated time series (default in days).
        N (int): The number of data points in the simulated time series.
        nLC (int, optional): Number of time series to simulate. Defaults to 1.
        log_flux (bool): Whether the flux/y values are in astronomical magnitude.
            This argument affects how errors are assigned. Defaults to True.
        lc_seed (int): Random seed for time series simulation. Defaults to None.

    Raises:
        RuntimeError: If the input CARMA term/model is not stable, thus cannot be
            solved by celerite.

    Returns:
        (array(float), array(float), array(float)): Time stamps (default in day), y
        values and measurement errors of the simulated time series.
    """

    assert isinstance(
        carmaTerm, celerite.celerite.terms.Term
    ), "carmaTerm must a celerite GP term"

    if (not isinstance(carmaTerm, DRW_term)) and (carmaTerm._arroots.real > 0).any():
        raise RuntimeError(
            "The covariance matrix of the provided CARMA term is not positive definite!"
        )

    if lc_seed is not None:
        rng = np.random.default_rng(seed=lc_seed)
    else:
        rng = np.random.default_rng()

    t = np.linspace(0, duration, N)
    noise = carmaTerm.get_rms_amp() / SNR
    yerr = rng.lognormal(0, 0.25, N) * noise
    yerr = yerr[np.argsort(np.abs(yerr))]  # small->large

    # init GP and solve matrix
    gp_sim = GP(carmaTerm)
    gp_sim.compute(t, yerr=1e-11)

    # simulate, assign yerr based on y
    t = np.repeat(t[None, :], nLC, axis=0)
    y = gp_sample(gp_sim, size=nLC, seed=lc_seed)

    # format yerr making it heteroscedastic
    yerr = np.repeat(yerr[None, :], nLC, axis=0)

    # if in mag, large value with large error; in flux, the opposite
    if log_flux:
        # ascending sort
        y_rank = y.argsort(axis=1).argsort(axis=1)
        yerr = np.array(list(map(lambda x, y: x[y], yerr, y_rank)))
    else:
        # descending sort
        y_rank = (-y).argsort(axis=1).argsort(axis=1)
        yerr = np.array(list(map(lambda x, y: x[y], yerr, y_rank)))

    if nLC == 1:
        return t[0], y[0], yerr[0]
    else:
        return t, y, yerr


def gpSimRand(
    carmaTerm,
    SNR,
    duration,
    N,
    nLC=1,
    log_flux=True,
    season=True,
    full_N=10_000,
    lc_seed=None,
    downsample_seed=None,
):
    """
    Simulate CARMA time series randomly downsampled from a much denser full time series.

    Args:
        carmaTerm (object): An EzTao CARMA kernel.
        SNR (float): Signal-to-noise defined as ratio between CARMA RMS amplitude and
            the median of the measurement errors (simulated using log normal).
        duration (float): The duration of the simulated time series (default in days).
        N (int): The number of data points in the simulated time series.
        nLC (int, optional): Number of time series to simulate. Defaults to 1.
        log_flux (bool): Whether the flux/y values are in astronomical magnitude.
            This argument affects how errors are assigned. Defaults to True.
        season (bool, optional): Whether to simulate 6-months seasonal gaps. Defaults
            to True.
        full_N (int, optional): The number of data points in the full time series
            (before downsampling). Defaults to 10_000.
        lc_seed (int): Random seed for full time series simulation. Defaults to None.
        downsample_seed (int): Random seed for downsampling the simulated full time
            series. Defaults to None.

    Returns:
        (array(float), array(float), array(float)): Time stamps (default in day), y
        values and measurement errors of the simulated time series.
    """
    t, y, yerr = gpSimFull(
        carmaTerm, SNR, duration, full_N, nLC=nLC, log_flux=log_flux, lc_seed=lc_seed
    )
    t = np.atleast_2d(t)
    y = np.atleast_2d(y)
    yerr = np.atleast_2d(yerr)

    # output t & yerr
    tOut = np.empty((nLC, N))
    yOut = np.empty((nLC, N))
    yerrOut = np.empty((nLC, N))

    # downsample
    for i in range(nLC):
        if season:
            mask1 = add_season(t[i])
        else:
            mask1 = np.ones(t[i].shape[0], dtype=bool)
        mask2 = downsample_byN(t[i, mask1], N, seed=downsample_seed)
        tOut[i, :] = t[i, mask1][mask2]
        yOut[i, :] = y[i, mask1][mask2]
        yerrOut[i, :] = yerr[i, mask1][mask2]

    if nLC == 1:
        return tOut[0], yOut[0], yerrOut[0]
    else:
        return tOut, yOut, yerrOut


def gpSimByTime(carmaTerm, SNR, t, nLC=1, log_flux=True, lc_seed=None):
    """
    Simulate CARMA time series at desired time stamps.

    Args:
        carmaTerm (object): An EzTao CARMA kernel.
        SNR (float): Signal-to-noise defined as ratio between CARMA RMS amplitude and
            the median of the measurement errors (simulated using log normal).
        t (array(float)): The desired time stamps (starting from zero).
        nLC (int, optional): Number of time series to simulate. Defaults to 1.
        log_flux (bool): Whether the flux/y values are in astronomical magnitude.
            This argument affects how errors are assigned. Defaults to True.
        lc_seed (int): Random seed for time series simulation. Defaults to None.

    Returns:
        (array(float), array(float), array(float)): Time stamps (default in day), y
        values and measurement errors of the simulated time series.

    Note: The simulated light curves will be different for different set of input times,
        even with the same lc_seed.
    """
    if lc_seed is not None:
        rng = np.random.default_rng(seed=lc_seed)
    else:
        rng = np.random.default_rng()

    t = np.atleast_1d(t)
    noise = carmaTerm.get_rms_amp() / SNR
    yerr = rng.lognormal(0, 0.25, len(t)) * noise
    yerr = yerr[np.argsort(np.abs(yerr))]  # small->large

    # init GP and solve matrix
    gp_sim = GP(carmaTerm)
    gp_sim.compute(t, yerr=1e-11)

    # simulate, assign yerr based on y
    t = np.repeat(t[None, :], nLC, axis=0)
    y = gp_sample(gp_sim, size=nLC, seed=lc_seed)

    # format yerr making it heteroscedastic
    yerr = np.repeat(yerr[None, :], nLC, axis=0)

    # if in mag, large value with large error; in flux, the opposite
    if log_flux:
        # ascending sort
        y_rank = y.argsort(axis=1).argsort(axis=1)
        yerr = np.array(list(map(lambda x, y: x[y], yerr, y_rank)))
    else:
        # descending sort
        y_rank = (-y).argsort(axis=1).argsort(axis=1)
        yerr = np.array(list(map(lambda x, y: x[y], yerr, y_rank)))

    if nLC == 1:
        return t[0], y[0], yerr[0]
    else:
        return t, y, yerr


def addNoise(y, yerr, seed=None):
    """
    Add (gaussian) noise to the input simulated time series given the measurement uncertainties.

    Args:
        y (array(float)): The 'clean' time series.
        yerr (array(float)): The measurement uncertainties for the input
            time series.
    seed (int): Random seed for simulating noise. Defaults to None.

    Returns:
        array(float): A new time series with simulated (gaussian ) noise added
        on top.
    """

    if seed is not None:
        rng = np.random.default_rng(seed=seed)
    else:
        rng = np.random.default_rng()

    vec_norm = np.vectorize(rng.normal, signature="(n),(n)->(n)")
    noise = vec_norm(np.zeros_like(y), yerr)

    return y + noise


def pred_lc(t, y, yerr, params, p, t_pred, return_var=True):
    """
    Generate predicted values at particular time stamps given the initial
    time series and a best-fit model.

    Args:
        t (array(float)): Time stamps of the initial time series.
        y (array(float)): y values (i.e., flux) of the initial time series.
        yerr (array(float)): Measurement errors of the initial time series.
        params (array(float)): Best-fit CARMA parameters
        p (int): The AR order (p) of the given best-fit model.
        t_pred (array(float)): Time stamps to generate predicted time series.
        return_var (bool, optional): Whether to return uncertainties in the mean
            prediction. Defaults to True.

    Returns:
        (array(float), array(float), array(float)): t_pred, mean prediction at t_pred
        and uncertainties (variance) of the mean prediction.
    """

    assert p >= len(params) - p, "The dimension of AR must be greater than that of MA"

    # get ar, ma
    ar = params[:p]
    ma = params[p:]

    # reposition lc
    y_aln = y - np.median(y)

    # init kernel, gp and compute matrix
    kernel = CARMA_term(np.log(ar), np.log(ma))
    gp = celerite.GP(kernel, mean=0)
    gp.compute(t, yerr)

    try:
        mu, var = gp.predict(y_aln, t_pred, return_var=return_var)
    except FloatingPointError as e:
        print(e)
        print("No (super small) variance will be returned")
        return_var = False
        mu, var = gp.predict(y_aln, t_pred, return_var=return_var)

    return t_pred, mu + np.median(y), var



"""
A set of functions to computer 2nd order statistics of CARMA models.
"""

import numpy as np
from .CARMATerm import acf, CARMA_term


def drw_psd(amp, tau):
    """
    Return a function that computes DRW Power Spectral Density (PSD).

    Args:
        amp (float): DRW RMS amplitude
        tau (float): DRW decorrelation/characteristic timescale

    Returns:
        A function that takes in frequencies and returns PSD at the given frequencies.
    """

    # convert amp, tau to CARMA parameters
    a0 = 1 / tau
    sigma2 = 2 * amp**2 * a0

    def psd(f):
        return sigma2 / (a0**2 + (2 * np.pi * f) ** 2)

    return psd


def carma_psd(arparams, maparams):
    """
    Return a function that computes CARMA Power Spectral Density (PSD).

    Args:
        arparams (array(float)): AR coefficients.
        maparams (array(float)): MA coefficients

    Returns:
        A function that takes in frequencies and returns PSD at the given frequencies.
    """
    arparams = np.insert(arparams, 0, 1)
    maparams = np.array(maparams)
    arparams_rv = arparams[::-1]

    def psd(f):
        # init terms
        num_terms = complex(0)
        denom_terms = complex(0)

        for i, param in enumerate(maparams):
            num_terms += param * np.power(2 * np.pi * f * (1j), i)

        for k, param in enumerate(arparams_rv):
            denom_terms += param * np.power(2 * np.pi * f * (1j), k)

        num = np.abs(np.power(num_terms, 2))
        denom = np.abs(np.power(denom_terms, 2))

        return num / denom

    return psd


def gp_psd(carmaTerm):
    """
    Return a function that computes native GP Power Spectral Density (PSD).

    Args:
        carmaTerm (object): A celerite CARMA term.
    Returns:
        A function that takes in frequencies and returns PSD at the given frequencies.
    """

    def psd(f):
        return 2.5 * carmaTerm.get_psd(2 * np.pi * f)

    return psd


def drw_acf(tau):
    """
    Return a function that computes the DRW autocorrelation function (ACF).

    Args:
        tau (float): DRW decorrelation/characteristic timescale.

    Returns:
        A function that takes in time lags and returns ACF at the given lags.
    """
    # convert to CARMA parameter
    a0 = 1 / tau

    def acf(lag):
        return np.exp(-a0 * lag)

    return acf


def carma_acf(arparams, maparams):
    """
    Return a function that computes the model autocorrelation function (ACF) of CARMA.

    Args:
        arparams (array(float)): AR coefficients.
        maparams (array(float)): MA coefficients.

    Returns:
        A function that takes in time lags and returns ACF at the given lags.
    """

    roots = np.roots(np.append([1], arparams)).astype(np.complex128)
    autocorr = acf(roots, arparams, maparams)
    gpTerm = CARMA_term(np.log(arparams), np.log(maparams))

    def autocorr_func(lag):
        R = 0
        for i, r in enumerate(roots):
            R += autocorr[i] * np.exp(r * lag)

        return np.real(R / gpTerm.get_rms_amp() ** 2)

    return autocorr_func


def drw_sf(amp, tau):
    """
    Return a function that computes the structure function (SF) of DRW.

    Args:
        amp (float): DRW RMS amplitude
        tau (float): DRW decorrelation/characteristic timescale.

    Returns:
        A function that takes in time lags and returns DRW SF at the given lags.
    """

    def sf(lag):
        return np.sqrt(2 * amp**2 * (1 - drw_acf(tau)(lag)))

    return sf


def carma_sf(arparams, maparams):
    """
    Return a function that computes the CARMA structure function (SF).

    Args:
        arparams (array(float)): AR coefficients.
        maparams (array(float)): MA coefficients.

    Returns:
        A function that takes in time lags and returns CARMA SF at the given lags.
    """
    gpTerm = CARMA_term(np.log(arparams), np.log(maparams))
    amp2 = gpTerm.get_rms_amp() ** 2

    def sf(lag):
        return np.sqrt(2 * amp2 * (1 - carma_acf(arparams, maparams)(lag)))

    return sf
